	; Memory soak tester by Simon Rowe (srowe@mose.org.uk)
	;
	; Based on C implementation in Michael Barr's article,
	;   "Fast Accurate Memory Test Suite"
	;   http://www.barrgroup.com/Embedded-Systems/How-To/Memory-Test-Suite-C

TIME	= $A0
PNT	= $D1
PNTR	= $D3
COLOR	= $0286
PUTSCRN	= $EAAA
COLORSYN = $EAB2
CHROUT	= $FFD2
STOP	= $FFE1
GETIN	= $FFE4
PLOT	= $FFF0

MEMSTART = $57
MEMLEN = $59
MEMEND = $5B
REPEAT = $5D
BITMASK = $5E
MEMCUR = $5F
PASSCNT = $61
FAILCNT = $62
	
*	= $0400

MAIN
	JSR STOP
	BEQ MAIN		; loop until STOP is up
	JSR SHOWFORM		; clear screen and display form
	LDA #$00
	STA PASSCNT
	STA FAILCNT
	LDX #$02
	LDY #$06
	JSR GOTO		; position for start address
	JSR INPUTHEX		; input start address
	BCC GETREST

EXIT1	LDX #$13		; if STOP pressed then exit
	LDY #$00
	JSR GOTO
	RTS
	
GETREST	STA MEMSTART		; save start addresss
	STY MEMSTART+1

	LDX #$02
	LDY #$12
	JSR GOTO		; position for length
	JSR INPUTHEX		; input region length
	BCS MAIN		; if STOP pressed then restart
	
	LDA MEMLEN		; save region length
	STY MEMLEN+1
	ORA MEMLEN+1
	BEQ EXIT1

	CLC			; Calculate end address
	LDA MEMLEN
	ADC MEMSTART
	STA MEMEND
	LDA MEMLEN+1
	ADC MEMSTART+1
	STA MEMEND+1

	JSR GETBITSRPT		; get repeat count & bit width
	BCS MAIN		; if STOP pressed then restart

	JSR SHOWPROG
AGAIN	JSR TESTDATA		; perform data bus test
	BCS FAIL
	LDA FAILCNT
	BNE ADMASK
	LDX #$08
	LDY #$0C
	JSR GOTO
	LDA #<PASS
	LDY #>PASS
	JSR PRINTSTR

ADMASK	JSR SETADDRMASK
	JSR ADDRHEADER
	JSR TESTADDR
	BCS FAIL
	LDA FAILCNT
	BNE DOCELL
	LDX #$0B
	LDY #$0C
	JSR GOTO
	LDA #<PASS
	LDY #>PASS
	JSR PRINTSTR

DOCELL	JSR CELLHEADER
	JSR CELLTEST
	BCS FAIL
	LDA FAILCNT
	BNE ALLDONE
	LDX #$0E
	LDY #$0C
	JSR GOTO
	LDA #<PASS
	LDY #>PASS
	JSR PRINTSTR

ALLDONE
	INC PASSCNT
	LDX #$05
	LDY #$07
	JSR GOTO
	LDA PASSCNT
	JSR PRINTDEC
	JMP CHKRPT

FAIL	INC FAILCNT
	LDX #$05
	LDY #$13
	JSR GOTO
	LDA FAILCNT
	JSR PRINTDEC
	
CHKRPT	LDA REPEAT
	BEQ AGAIN		; continuous test
	CLC
	LDA PASSCNT
	ADC FAILCNT
	CMP REPEAT
	BCC AGAIN
	
EXIT	LDX #$13
	LDY #$00
	JSR GOTO
	RTS

GETBITSRPT
	LDX #$03
	LDY #$06
	JSR GOTO		; position for repeat count
	SEC
	JSR INPUTDEC		; input repeat count
	BCS GBSOUT
	STA REPEAT
GETBITS	LDX #$03
	LDY #$12
	JSR GOTO		; position for bit width
	CLC
	JSR INPUTDEC		; input bit width
	BCS GBSOUT
	CMP #$04
	BEQ BITSOK
	CMP #$08
	BNE GETBITS		; repeat if invalid width
	LDY #$FF
	.BYTE $2C		; makes next line BIT $0FA5
BITSOK	LDY #$0F
	STY BITMASK
	CLC
GBSOUT	RTS


;***********************************************************************************;
;
; display banner and initial fields

SHOWFORM
	LDA #<BANNER
	LDY #>BANNER
	JSR PRINTSTR
	LDX #$02
	LDY #$00
	JSR GOTO
	LDA #<FORM1
	LDY #>FORM1
	JSR PRINTSTR
	LDX #$03
	LDY #$00
	JSR GOTO
	LDA #<FORM2
	LDY #>FORM2
	JMP PRINTSTR


;***********************************************************************************;
;
; display pass/fail and data bus test header

SHOWPROG
	LDX #$05
	LDY #$01
	JSR GOTO		; position for pass/fail counters
	LDA #<PASSFAIL
	LDY #>PASSFAIL
	JSR PRINTSTR
	LDX #$07
	LDY #$01
	JSR GOTO		; position for data bus test
	LDA #<DATABUS
	LDY #>DATABUS
	JMP PRINTSTR


;***********************************************************************************;
;
; display address bus test header

ADDRHEADER
	LDX #$0A
	LDY #$01
	JSR GOTO		; position for data bus test
	LDA #<ADDRBUS
	LDY #>ADDRBUS
	JMP PRINTSTR


;***********************************************************************************;
;
; display cell test header

CELLHEADER
	LDX #$0D
	LDY #$01
	JSR GOTO		; position for cell test
	LDA #<CELLT
	LDY #>CELLT
	JMP PRINTSTR


;***********************************************************************************;
;
; Test the data bus wiring in a memory region by
; performing a walking 1's test at the start of every page.
;
; MEMSTART - start address
; MEMEND - end address
;
; Returns - Cb = 0 - success

REFVAL	= $63

TESTDATA
	LDA MEMSTART		; Reset current pointer
	STA MEMCUR
	LDA MEMSTART+1
	STA MEMCUR+1
TDLOOP1 LDX #$07
	LDY #$0C
	JSR GOTO
	LDA MEMCUR		; Display current address
	LDY MEMCUR+1
	JSR PRINTHEX
	LDY #$00
	LDA #$01		; Perform a walking 1's test at the given address
TDLOOP2	STA (MEMCUR),Y		; Write the test pattern
	STA REFVAL
	LDA (MEMCUR),Y
	AND BITMASK
	CMP REFVAL
	BNE TDFAIL
	ASL
	AND BITMASK
	BNE TDLOOP2
	LDA #5
	JSR SLEEP

	INC MEMCUR+1		; Move to next page
	LDA MEMCUR+1
	CMP MEMEND+1
	BCC TDLOOP1
TDDONE	CLC
	RTS
TDFAIL	PHA			; Report details of failure
	LDX #$08
	LDY #$02
	JSR GOTO
	LDA REFVAL
	JSR PRINTBIN
	LDX #$08
	LDY #$0C
	JSR GOTO
	LDX COLOR
	LDA #$02
	STA COLOR
	PLA
	JSR PRINTBIN
	TXA
	STA COLOR
	SEC
	RTS


;***********************************************************************************;
;
; set the address mask based on the most significant bit of the region length

ADDRMASK = $64

SETADDRMASK
	LDA #$80		; start bit mask as %10000000 00000000
	STA ADDRMASK+1
	LDA #$00
	STA ADDRMASK
AMLOOP	LDA MEMLEN+1
	AND ADDRMASK+1
	BNE AMDONE		; if bit set in region length then stop
	LDA MEMLEN
	AND ADDRMASK
	BNE AMDONE
	LSR ADDRMASK+1		; shift bit mask right
	ROR ADDRMASK
	JMP AMLOOP

AMDONE	LDA ADDRMASK		; convert to address mask
	BNE AMDEC
	DEC ADDRMASK+1
AMDEC	DEC ADDRMASK
	RTS


;***********************************************************************************;
;
; Test the address bus wiring in a memory region by
; performing a walking 1's test on the relevant bits
; of the address and checking for aliasing. This test
; will find single-bit address failures such as stuck
; -high, stuck-low, and shorted pins.
;
; MEMSTRT - start address
; MEMLEN - region length
;
; Returns - Cb = 0 - success

ADDRBIT = $66
ERRMSG = $68

TESTADDR
	LDY #$01		; initialize address bit
	STY ADDRBIT
	DEY
	STY ADDRBIT+1
	
	; write the sentinel at each power of 2 address
	
TALOOP1	LDA ADDRBIT		; compare with bitmask
	AND ADDRMASK
	BNE TACHKS1
	LDA ADDRBIT+1
	AND ADDRMASK+1
	BEQ TACONT1		; bit now beyond bitmask
TACHKS1	LDA ADDRBIT		; compare with start address
	AND MEMSTART
	BNE TACONT1		; bit now into start address
	LDA ADDRBIT+1
	AND MEMSTART+1
	BNE TACONT1		; bit now into start address

	LDA MEMSTART		; combine to form pointer
	ORA ADDRBIT
	STA MEMCUR
	LDA MEMSTART+1
	ORA ADDRBIT+1
	STA MEMCUR+1

	LDA #$55
	STA (MEMCUR),Y		; store sentinel value

	ASL ADDRBIT
	ROL ADDRBIT+1
	JMP TALOOP1

TACONT1	LDA #$AA		; store inverse sentinel at start
	STA (MEMSTART),Y

	LDY #$01		; initialize address bit
	STY ADDRBIT
	DEY
	STY ADDRBIT+1

	; check the sentinel at each power of 2 address, if it has changed
	; an address line is stuck high
	
TALOOP2	LDA ADDRBIT		; compare with bitmask
	AND ADDRMASK
	BNE TACHKS2
	LDA ADDRBIT+1
	AND ADDRMASK+1
	BEQ TACONT2		; bit now beyond bitmask
TACHKS2	LDA ADDRBIT		; compare with start address
	AND MEMSTART
	BNE TACONT2		; bit now into start address
	LDA ADDRBIT+1
	AND MEMSTART+1
	BNE TACONT2		; bit now into start address

	LDA MEMSTART		; combine to form pointer
	ORA ADDRBIT
	STA MEMCUR
	LDA MEMSTART+1
	ORA ADDRBIT+1
	STA MEMCUR+1

	LDX #$0A		; display current address
	LDY #$0C
	JSR GOTO
	LDA MEMCUR
	LDY MEMCUR+1
	JSR PRINTHEX
	LDA #5
	JSR SLEEP

	LDY #$00
	LDA (MEMCUR),Y		; fetch sentinel value
	EOR #$55		; invert
	AND BITMASK
	BEQ TASHFT2
	LDA #<STUCKHI
	STA ERRMSG
	LDA #>STUCKHI
	STA ERRMSG+1
	JMP TAFAIL

TASHFT2	ASL ADDRBIT
	ROL ADDRBIT+1
	JMP TALOOP2

TACONT2	LDA #$55
	STA (MEMSTART),Y

	LDY #$01		; initialize address bit
	STY ADDRBIT
	DEY
	STY ADDRBIT+1

	; write the inverse sentinel at each power of 2 address
	
TALOOP3	LDA ADDRBIT		; compare with bitmask
	AND ADDRMASK
	BNE TACHKS3
	LDA ADDRBIT+1
	AND ADDRMASK+1
	BEQ TACONT3		; bit now beyond bitmask
TACHKS3	LDA ADDRBIT		; compare with start address
	AND MEMSTART
	BNE TACONT3		; bit now into start address
	LDA ADDRBIT+1
	AND MEMSTART+1
	BNE TACONT3		; bit now into start address

	LDA MEMSTART		; combine to form pointer
	ORA ADDRBIT
	STA MEMCUR
	LDA MEMSTART+1
	ORA ADDRBIT+1
	STA MEMCUR+1
	LDA #$AA
	STA (MEMCUR),Y		; store inverse sentinel value

	LDX #$0A		; display current address
	LDY #$0C
	JSR GOTO
	LDA MEMCUR
	LDY MEMCUR+1
	JSR PRINTHEX
	LDA #5
	JSR SLEEP

	LDY #$00
	LDA (MEMSTART),Y
	EOR #$55
	AND BITMASK
	BEQ TASHORT
	LDA #<STUCKLO
	STA ERRMSG
	LDA #>STUCKLO
	STA ERRMSG+1
	JMP TAFAIL

TASHORT	JSR TESTADDR2
	BCS TAFAIL

	LDA #$55
	STA (MEMCUR),Y		; restore sentinel value

	ASL ADDRBIT
	ROL ADDRBIT+1
	JMP TALOOP3

TACONT3	CLC
	RTS

TAFAIL	LDX #$0C		; report failure cause
	LDY #$06
	JSR GOTO
	LDA COLOR
	PHA
	LDA #$02
	STA COLOR
	LDA #ERRMSG
	LDY #ERRMSG+1
	JSR PRINTSTR
	PLA
	STA COLOR
	SEC
	RTS


;***********************************************************************************;
;
; inner loop of address bus test

ADDRBIT2 = $6A
MEMCUR2 = $6C

TESTADDR2
	LDY #$01		; initialize address bit
	STY ADDRBIT2
	DEY
	STY ADDRBIT2+1

A2LOOP	LDA ADDRBIT2		; compare with bitmask
	AND ADDRMASK
	BNE A2CHKS
	LDA ADDRBIT2+1
	AND ADDRMASK+1
	BEQ A2DONE		; bit now beyond bitmask
A2CHKS	LDA ADDRBIT2		; compare with start address
	AND MEMSTART
	BNE A2DONE		; bit now into start address
	LDA ADDRBIT2+1
	AND MEMSTART+1
	BNE A2DONE		; bit now into start address

	LDA MEMSTART		; combine to form pointer
	ORA ADDRBIT2
	STA MEMCUR2
	LDA MEMSTART+1
	ORA ADDRBIT2+1
	STA MEMCUR2+1

	LDA (MEMCUR2),Y
	EOR #$55
	AND BITMASK
	BEQ A2SHFT
	LDA MEMCUR2		; compare both loop pointers
	EOR MEMCUR
	BNE A2ERR
	LDA MEMCUR2+1
	EOR MEMCUR+1
	BEQ A2SHFT		; pointers equal, no error

A2ERR	LDA #<SHORTED
	LDY #>SHORTED
	SEC
	RTS

A2SHFT	ASL ADDRBIT2
	ROL ADDRBIT2+1
	JMP A2LOOP

A2DONE	CLC
	RTS


;***********************************************************************************;
;
; Test the integrity of a physical memory device by
; performing an increment/decrement test over the
; entire region.  In the process every storage bit
; in the device is tested as a zero and a one.
;
; MEMSTART - start address
; MEMEND - end address
;
; Returns - Cb = 0 - success

ERRVAL = $6E

CELLTEST
	LDY #$00		; initialize fill value
	STY REFVAL

	LDA MEMSTART		; set pointer to start
	STA MEMCUR
	LDA MEMSTART+1
	STA MEMCUR+1

CTLOOP1	LDA MEMCUR		; compare with end address
	EOR MEMEND
	BNE CTFILL
	LDA MEMCUR+1
	EOR MEMEND+1
	BEQ CTCONT1

CTFILL	INC REFVAL		; preincrement fill value
	LDA REFVAL
	STA (MEMCUR),Y		; store fill value

	INC MEMCUR
	BNE CTLOOP1
	INC MEMCUR+1
	JMP CTLOOP1

CTCONT1	LDY #$00		; initialize fill value
	STY REFVAL

	LDA MEMSTART		; set pointer to start
	STA MEMCUR
	LDA MEMSTART+1
	STA MEMCUR+1

CTLOOP2	LDA MEMCUR		; compare with end address
	EOR MEMEND
	BNE CTCHK2
	LDA MEMCUR+1
	EOR MEMEND+1
	BEQ CTCONT2

CTCHK2	LDX #$0D		; display current address
	LDY #$0C
	JSR GOTO
	LDA MEMCUR
	LDY MEMCUR+1
	JSR PRINTHEX

	LDY #$00
	INC REFVAL		; preincrement fill value
	LDA (MEMCUR),Y
	STA ERRVAL
	EOR REFVAL		; compare with fill value
	AND BITMASK
	BNE CTFAIL

	LDA ERRVAL		; invert fill value
	EOR #$FF
	STA (MEMCUR),Y

	INC MEMCUR
	BNE CTLOOP2
	INC MEMCUR+1
	JMP CTLOOP2

CTCONT2	LDY #$00		; initialize fill value
	STY REFVAL

	LDA MEMSTART		; set pointer to start
	STA MEMCUR
	LDA MEMSTART+1
	STA MEMCUR+1

CTLOOP3	LDA MEMCUR		; compare with end address
	EOR MEMEND
	BNE CTCHK3
	LDA MEMCUR+1
	EOR MEMEND+1
	BEQ CTDONE

CTCHK3	LDX #$0D		; display current address
	LDY #$0C
	JSR GOTO
	LDA MEMCUR
	LDY MEMCUR+1
	JSR PRINTHEX

	LDY #$00
	INC REFVAL		; preincrement fill value
	LDA (MEMCUR),Y
	STA ERRVAL
	EOR REFVAL		; compare with fill value
	AND BITMASK
	CMP BITMASK
	BNE CTFAIL

	INC MEMCUR
	BNE CTLOOP3
	INC MEMCUR+1
	JMP CTLOOP3

CTDONE	CLC
	RTS

CTFAIL	LDX #$0E
	LDY #$02
	JSR GOTO
	LDA REFVAL
	JSR PRINTBIN
	LDX #$0E
	LDY #$0C
	JSR GOTO
	LDX COLOR
	LDA #$02
	STA COLOR
	LDA ERRVAL
	JSR PRINTBIN
	TXA
	STA COLOR
	SEC
	RTS


;***********************************************************************************;
;
; input 16 bit hex data field


LEN = $6F
VAL = $70

INPUTHEX
	LDY #$04
	STY LEN
	LDA #$00		; clear return value
	STA VAL
	STA VAL+1
INLOOP	LDY PNTR		; set reverse video on current char
	LDA (PNT),Y
	ORA #$80
	STA (PNT),Y
INKEY	JSR STOP
	BNE INCONT
	LDY PNTR
	LDA (PNT),Y
	AND #$7F
	STA (PNT),Y
	SEC
	RTS
INCONT	JSR GETIN
	BEQ INKEY
	TAX
	CMP #'0'		; less than numeric
	BCC INKEY
	CMP #'F'+1		; not a hex digit
	BCS INKEY
	CMP #'9'+1
	BCC INNIB		; decimal digit
	SBC #'A'-'9'-1
	BMI INKEY		; not a hex digit
INNIB	AND #$0F
	ASL VAL
	ROL VAL+1
	ASL VAL
	ROL VAL+1
	ASL VAL
	ROL VAL+1
	ASL VAL
	ROL VAL+1
	ORA VAL
	STA VAL
	TXA
	JSR CHROUT
	DEC LEN
	BNE INLOOP
	LDA VAL
	LDY VAL+1
	CLC
INOUT	RTS
	

;***********************************************************************************;
;
; input decimal data field
;
;  Cb = 0 - 1 digit
;  Cb = 1 - 2 digit

INPUTDEC
	LDY #$00
	STY VAL
	INY
	STY LEN
	BCC DCLOOP
	INC LEN
DCLOOP	LDY PNTR
	LDA (PNT),Y
	ORA #$80
	STA (PNT),Y
DCKEY	JSR STOP
	BNE DCCONT
	LDY PNTR
	LDA (PNT),Y
	AND #$7F
	STA (PNT),Y
	SEC
	RTS
DCCONT	JSR GETIN
	BEQ DCKEY
	TAX
	CMP #'0'
	BCC DCKEY
	CMP #'9'+1
	BCS DCKEY
	AND #$0F
	ASL VAL
	LDY VAL
	STY VAL+1
	ASL VAL
	ASL VAL
	CLC
	ADC VAL+1
	ADC VAL
	STA VAL
	TXA
	JSR CHROUT
	DEC LEN
	BNE DCLOOP
	LDA VAL
	CLC
	RTS


;***********************************************************************************;
;
; display character to screen at current cursor position

PRINTAT
	PHA			; save character
	JSR COLORSYN		; calculate pointer to colour RAM
	PLA			; restore character
	LDX COLOR		; get current colour code
	JMP PUTSCRN		; save character and colour to screen @ cursor

GOTO
	CLC
	JMP PLOT		; set X,Y cursor position


;***********************************************************************************;
;
; print a NUL terminated string
;
; .A - LSB
; .Y - MSB

STRPTR = $6F

PRINTSTR
	STA STRPTR
	STY STRPTR+1
	LDY #$00
	LDA (STRPTR),Y
PSLOOP	JSR CHROUT
	INY
	LDA (STRPTR),Y
	BNE PSLOOP
	RTS


;***********************************************************************************;
;
; Print a 16-bit hex value.
;
; .A - LSB
; .Y - MSB

PRINTHEX
	PHA
	TYA
	LSR
	LSR
	LSR
	LSR
	JSR PRTNIB
	TYA
	AND #$0F
	JSR PRTNIB
	PLA
	PHA
	LSR
	LSR
	LSR
	LSR
	JSR PRTNIB
	PLA
	AND #$0F
	JMP PRTNIB

PRTNIB	CMP #10
	BCS PGT10
	CLC
	ADC #"0"
	JMP CHROUT
PGT10	CLC
	ADC #"A"-10
	JMP CHROUT


;***********************************************************************************;
;
; print 8-bit binary value in .A

PRINTBIN
	LDY #$08
	PHA
PBLOOP	PLA
	ASL
	PHA
	BCS PB1
	LDA #"0"
	.byte $2C
PB1	LDA #"1"
	JSR CHROUT
	DEY
	BNE PBLOOP
	PLA
	RTS


;***********************************************************************************;
;
; print 8-bit decimal value in .A

TENS = $6F
UNITS = $70
DVAL = $71

PRINTDEC
	AND #$7F
	STA DVAL
	LDA #$00
	STA TENS
	STA UNITS
	LDY #$08
PDLOOP	ASL DVAL
	ROL UNITS
	LDA UNITS
	CMP #10
	BCC PDOVER1
	SBC #10
	STA UNITS
PDOVER1	ROL TENS
	DEY
	BNE PDLOOP
	LDA TENS
	CMP #10
	BCC PDOVER2
	SBC #10
PDOVER2	ORA #'0'
	JSR CHROUT
	LDA UNITS
	ORA #'0'
	JMP CHROUT


;***********************************************************************************;
;
; delay for .A jiffies

SLEEP
	CLC
	ADC TIME+2
SLOOP	CMP TIME+2
	BNE SLOOP
	RTS

BANNER
	.BYTE 147, "** MEMORY SOAK TEST **",$00
FORM1
	.BYTE "START:0000 LENGTH:0000",$00
FORM2
	.BYTE "RUNS: 00   WIDTH: 8",$00
PASSFAIL
	.BYTE "PASS: 00    FAIL: 00",$00
DATABUS
	.BYTE "DATA BUS:  0000",$00
ADDRBUS
	.BYTE "ADDR BUS:  0000",$00
CELLT
	.BYTE "CELL TEST: 0000",$00
PASS
	.BYTE "PASS",$00
STUCKHI
	.BYTE $12,"STUCK HIGH",$92,$00
STUCKLO
	.BYTE $12,"STUCK LOW",$92,$00
SHORTED
	.BYTE $12,"SHORTED",$92,$00
